; BOOTLOADER V2.0
; general purpose bootloader for FAT12 and FAT16 filesystems, supports loading from partitioned and unpartitioned disks

[org 0xFc24]		; code offset

; initialize data segment and stack
xor ax, ax
mov ds, ax
mov es, ax

; THIS DOES NOT WORK WITH SOME BIOS IMPLEMENTATIONS!
; The DL register SHOULD have the physical drive number of the boot device
; but in some BIOSes (at least in Nokia MikroMikko BIOS) it has always 0x00
%ifdef NO_MIKROMIKKO_FIX
mov byte [0x7C15], dl ; save the boot drive
%endif

; move to the end of the segment
mov si, 0x7c00
mov di, 0xFc00
mov cx, 0x100
cld
rep movsw

; dirty hack that breaks booting from B or D,E,F... drives :(
%ifndef NO_MIKROMIKKO_FIX
;cmp byte [0xFC15], 0xF8
;jne floppy
;mov al, 0x80
;floppy:
;mov byte [0xFC15], al
;xor al, al

xchg byte [0xFC15], al ; 0x00 to 0xFC15
xor al, 0xF8           ; result zero if booting from hard disk
jnz floppy
mov byte [0xFC15], 0x80
floppy:
%endif

;mov bx, start
;jmp bx
db 0xE9 ; relative jump
dw 0x8000

start:
mov ss, cx

retry:
mov sp, 0x400

; read the first sector of root directory into 0000:0600
xor ch, ch
mov cl, [0xFC10] ; FAT copies
xor ax, ax
xor dx, dx
fat_loop1:
add ax, [0xFC16] ; sectors per FAT
adc dx, 0
loop fat_loop1

; now DX:AX = FIRST ROOTDIR SECTOR

push ax
push dx
mov bx, [0xFC11] ; BX = ROOT_DIR_ENTRIES
mov cl, 4
shr bx, cl
add ax, bx
adc dx, 0
; now DX:AX = first data sector
mov [0xFC02], ax
mov [0xFC04], dx

; now CLUSTERS_IN_FS = (SECTORS_IN_FS - DX:AX) / SECTORS_PER_CLUSTER
; if [0xFC13] (16-bit SECTORS_IN_FS) == 0, the filesystem must be FAT16
; else if (CLUSTERS_IN_FS+2)/256 > SECTORS_PER_FAT, the filesystem is FAT12

xor dx, dx
mov ax, [0xFC13] ; ax = SECTORS_IN_FS
sub ax, [0xFC02] ; ax = ax - FIRST_DATA_SECTOR, now ax = CLUSTERS_IN_FS * SECTORS_PER_CLUSTER
jc large_harddisk
mov bl, [0xFC0D]
xor bh, bh
div bx             ; ax = ax / SECTORS_PER_CLUSTER, now AX = CLUSTERS_IN_FS
add ax, 2

mov al, ah
xor ah, ah        ; ax == (CLUSTERS_IN_FS+2)/256

mov [0xFC13], ax

;jc large_harddisk

mov ax, [0xFC16] ; ax = SECTORS_PER_FAT

sub [0xFC13], ax  ; if [0xFC13] & 0x8000, the filesystem is FAT16
jmp floppy_or_hd_end
large_harddisk:
or byte [0xFC14], 0x80
floppy_or_hd_end:

pop dx
pop ax

; number of root directory entries
mov cx, [0xFc11]
read_root_sector:
mov bx, 0x600
mov si, bx
call read_sector
root_loop:

push cx

mov si, bx
mov di, kernelfile
mov cx, 11
repe cmpsb
  
pop cx

jz found_kernel

dec cx
jz nokernelfound

add bx, 32
cmp bh, 0x08
je next_root_sector
jmp root_loop
next_root_sector:
call long_inc
jmp read_root_sector

long_inc:
;add ax, 1
;adc dx, 0
inc ax
jnz long_inc_ret
inc dx
long_inc_ret:
ret


found_kernel:
; bx == pointer to directory entry of the kernel file
; bx + 28 == file size
; bx + 26 == starting cluster
;mov si, kernelfound
;call puts

;xor ch, ch

mov ax, [bx+26] ; ax = starting cluster of the kernel file

; load kernel to 0x600

mov bx, 0x600
read_kernel_loop:
call read_cluster

call next_cluster

cmp ax, 0xFFF7
jb read_kernel_loop


; follow kernel file's cluster chain and load the file to memory...

; jump to kernel
jmp 0x60:0


nokernelfound:
call errormsg
halt:
jmp halt

; ax == cluster
; returns: cluster value to ax
next_cluster:
  push bx
  ;push cx

  test al, 1
  pushf

  mov bx, ax

  ;test byte [0xFC15], 0x80 ; are we loading from a hard disk? 
  ;jne harddisk1
  test byte [0xFC14], 0x80
  jne harddisk1

; if not, offset to the cluster: cluster * 1,5
  shr ax, 1

  harddisk1:
; else offset to the cluster: cluster * 2
  xor dl, dl
  add bx, ax
; store the more significant bits of the offset to dl
  rcr dl, 1

; bx = offset to the cluster

  mov ax, bx
  ; bx = bx % 512
  and bx, 0x01FF

; save the offset to stack
  push bx

; divide with 512 (sector size)
  mov cl, 9
  shr ax, cl
; and keep the more significant bit while dividing
  or al, dl
  xor dl, dl
  

  mov cl, 1
  cmp bx, 0x1FF
  jne read_one_sector
  inc cl
  read_one_sector:

  ; --- <caching> ---
  cmp ax, [0xFFFE] ; are the required FAT sectors already in memory?
  jne not_in_cache
  cmp dl, [0xFC01]
  jne not_in_cache
  cmp cl, [0xFC00]
  je read_cluster_table_end

  not_in_cache:
  mov [0xFFFE], ax
  mov [0xFC00], cl
  mov [0xFC01], dl
  ; --- </caching> ---

; read one or two sectors from sector ax
  mov bx, 0xF800
  read_cluster_table:
  call read_sector
  dec cl
  jz read_cluster_table_end
  add bh, 0x02
  call long_inc
; read also the next sector
  jmp read_cluster_table
  read_cluster_table_end:
  pop bx

  ; bx += 0xF800
  add bx, 0xF800

  mov ax, [bx]

  ;test byte [0xFC15], 0x80 ; are we loading from a hard disk?
  ;jne harddisk2
  test byte [0xFC14], 0x80
  jne harddisk2

  popf

  jnz next_cluster_odd
  and ax, 0xFFF
  jmp next_cluster_end

  next_cluster_odd:
  mov cl, 4
  shr ax, cl

  next_cluster_end:

  cmp ax, 0xFF7
  jl next_cluster_end_hd
  or ah, 0xF0

  next_cluster_end_hd:
  ;pop cx
  pop bx
ret
harddisk2:
popf
jmp next_cluster_end_hd

; reads cluster
; ax == cluster to read
; es:bx == pointer to buffer
; increments bx by the amount of read bytes
read_cluster:
  push ax
  
  ; sector to read = (cluster-2) * sectors per cluster
  xor cx, cx
  mov dx, cx
  mov cl, [0xFC0D]
  sub ax, 2
  mul cx

  add ax, [0xFC02]
  adc dx, [0xFC04]
  read_cluster_loop:
  call read_sector
  add bh, 0x02

  ; don't overwrite the bootloader...
  cmp bh, 0xF8
  jnb read_cluster_end
  
  call long_inc
  loop read_cluster_loop

  read_cluster_end:
  pop ax
ret

; reads logical sector from 1,44 MB floppy
; dx:ax == sector
; es:bx = pointer to buffer
read_sector:
  push ax
  push cx
  push dx
  ;push si
  mov si, 16 ; number of retries
  
  call long_inc ; reserved sectors
  add ax, [0xFC1C] ; hidden sectors
  adc dx, [0xFC1E]

  ; "sector offset" == sectors per cluster * cluster
  ; now ax == sector offset

  ; track number: sector offset / sectors per track
  ; 1,44 MB diskette has 18 sectors per track
  div word [0xFc18]

  ; now ax == track number
  ; dx == remainder == sector to read
  mov cl, dl
  inc cl

  divide_track_number:
  ; divide track number with number of heads
  xor dx, dx
  div word [0xFc1A]
  mov ch, al

  ror ah, 1
  ror ah, 1
  or cl, ah

  read_sector_head:
  ; remainder == head...
  mov dh, dl

  ; drive number: 0 (floppy), 80h: hard disk
  mov dl, byte [0xFC15]
  read_floppy:

;  write dot
  push bx

  mov bx, 0x0007
  mov ax, 0x0E2E
  int 0x10

  pop bx

  read_sector_retry:
  mov ax, 0x0201 ; read one sector
  int 0x13
  jnc read_sector_end

  xor ah, ah
  int 0x13

  dec si
  jnz read_sector_retry

  call errormsg
  jmp retry

  read_sector_end:

  ;pop si
  pop dx
  pop cx
  pop ax
ret

; si == pointer to string
errormsg:

  ;mov si, error
  ;mov bx, 0x0007
  ;mov ah, 0x0E

  ;puts_loop:
  ;mov al, [si]
  ;int 0x10
  ;inc si
  ;cmp al, 0x0A
  ;je puts_end
  ;jmp puts_loop
  ;puts_end:
  mov ax, 0x0E00 + 'E'
  int 0x10
  
ret

kernelfile: db "KERNEL  BIN"
;error: db "ERR", 0x0A

times 0x01FE-0x24-($-$$) db 0

dw 0xAA55
